package com.css.simulation.resource.scheduler.util;

import api.common.pojo.common.ResponseBodyVO;
import api.common.pojo.constants.DictConstants;
import api.common.pojo.dto.ProjectMessageDTO;
import api.common.util.*;
import com.css.simulation.resource.scheduler.configuration.custom.CustomConfiguration;
import com.css.simulation.resource.scheduler.configuration.feign.WebServerClient;
import com.css.simulation.resource.scheduler.configuration.kubernetes.KubernetesConfiguration;
import com.css.simulation.resource.scheduler.configuration.redis.CustomRedisClient;
import com.css.simulation.resource.scheduler.data.entity.NodeEntity;
import com.css.simulation.resource.scheduler.data.entity.PrefixEntity;
import com.css.simulation.resource.scheduler.data.entity.ProjectEntity;
import com.css.simulation.resource.scheduler.data.entity.UserEntity;
import com.css.simulation.resource.scheduler.data.model.NodeModel;
import com.css.simulation.resource.scheduler.mapper.AutoSubProjectMapper;
import com.css.simulation.resource.scheduler.mapper.ClusterMapper;
import com.css.simulation.resource.scheduler.mapper.ManualProjectMapper;
import com.css.simulation.resource.scheduler.mapper.UserMapper;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.kubernetes.client.openapi.ApiClient;
import io.kubernetes.client.openapi.ApiException;
import lombok.SneakyThrows;
import lombok.Synchronized;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.impl.client.CloseableHttpClient;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.scheduling.annotation.Async;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.io.File;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.stream.Collectors;

/**
 * 任务结果打分
 */
@Component
@Slf4j
public class ProjectUtil {

    @Value("${scheduler.linux-path.pod-yaml-directory}")
    private String podYamlDirectory;
    @Resource
    private ManualProjectMapper manualProjectMapper;
    @Resource
    private AutoSubProjectMapper autoSubProjectMapper;
    @Resource
    private UserMapper userMapper;
    @Resource
    private ClusterMapper clusterMapper;
    @Resource
    private KubernetesConfiguration kubernetesConfiguration;
    @Resource
    private ApiClient apiClient;
    @Resource
    private StringRedisTemplate stringRedisTemplate;
    @Resource
    private CustomRedisClient customRedisClient;
    @Resource
    private WebServerClient webServerClient;
    @Resource
    private CustomConfiguration customConfiguration;
    @Resource
    private CloseableHttpClient closeableHttpClient;
    @Resource
    private RequestConfig requestConfig;

    public String getIsChoiceGpuByProjectId(String projectId) {
        return getProjectByProjectId(projectId).getIsChoiceGpu();
    }


    @SneakyThrows
    public void deleteYamlByProjectId(String projectId) {
        List<String> absolutePathList = FileUtil.listAbsolutePath(podYamlDirectory);
        for (String absolutePath : absolutePathList) {
            if (absolutePath.contains(projectId)) {
                boolean delete = new File(absolutePath).delete();
                log.debug("删除结果：" + delete);
            }
        }
    }


    public List<NodeEntity> getNodeListToCount(Map<String, Integer> nodeMap) {
        List<NodeEntity> result = new ArrayList<>();
        nodeMap.forEach((nodeName, parallelism) -> {
            for (int i = 0; i < parallelism; i++) {
                result.add(new NodeEntity(nodeName, 0));
            }
        });
        return result;
    }

    /**
     * 判断算法是否已经导入
     *
     * @return 算法是否已经导入
     */
    @SneakyThrows
    public boolean isImported(String dockerImageWithoutVersion) {
        String dockerImageListString = LinuxUtil.execute("docker images");
        return dockerImageListString.contains(dockerImageWithoutVersion);
    }

    public String getRandomPodName(String projectId) {
        return "project-" + projectId + "-" + StringUtil.getRandomEightBitUUID();
    }

    public String getPodYamlName(String nodeName, String podName) {
        return nodeName + "#" + podName + ".yaml";
    }


    public void deletePod(String podName) {
        try {
            // 先删除 redis key
            KubernetesUtil.deletePod(apiClient, kubernetesConfiguration.getNamespace(), podName);
            log.info("等待 pod " + podName + " 的资源释放完成。");
            TimeUnit.SECONDS.sleep(7);
        } catch (ApiException apiException) {
            log.info("pod " + podName + " 已删除。");
        } catch (Exception e) {
            e.printStackTrace();
            log.error("删除 pod " + podName + " 报错。", e);
        }
    }


    public String getNodeNameOfPod(String projectId, String podName) {
        String key = "project:" + projectId + ":pod:" + podName + ":node";
        final String s = stringRedisTemplate.opsForValue().get(key);
        if (StringUtil.isEmpty(s)) {
            throw new RuntimeException("无法获取 pod 运行所在节点：" + key);
        }
        return s;
    }


    /**
     * 更改一个名字继续启动
     *
     * @param projectId   项目 id
     * @param nodeName    运行 pod 的节点名称
     * @param lastPodName 即将删除的 pod 名称
     */
    @SneakyThrows
    public void createNextPod(String userId, String projectId, String projectType, String nodeName, String lastPodName) {
        final String isChoiceGpu = getIsChoiceGpuByProjectId(projectId);
        log.info("删除上一个 pod：projectId={}，nodeName={}，lastPodName={}", projectId, nodeName, lastPodName);
        String cpuOrderString = stringRedisTemplate.opsForValue().get("project:" + projectId + ":pod:" + lastPodName + ":cpu");
        deletePod(lastPodName);
        RedisUtil.deleteByKey(stringRedisTemplate, "project:" + projectId + ":pod:" + lastPodName + ":cpu");
        //2 获取新的 yaml 信息
        final Set<String> yamlPathCacheKeySet = RedisUtil.getKeySetByPrefix(stringRedisTemplate, "project:" + projectId + ":node:" + nodeName + ":yaml");
        if (CollectionUtil.isEmpty(yamlPathCacheKeySet)) {
            // 如果当前节点没有下一个yaml，则返回一个并行度。
            log.info("节点 " + nodeName + " 已经执行完被分配的项目 " + projectId + " 的所有 pod。");
            incrementOneParallelism(isChoiceGpu, nodeName);
            releaseLicense(userId, getModelTypeByProjectIdAndProjectType(projectId, projectType), 1);
        } else {
            final String yamlPathCacheKey = new ArrayList<>(yamlPathCacheKeySet).get(0);
            final String absolutePath = stringRedisTemplate.opsForValue().get(yamlPathCacheKey);
            // 修改 cpu 编号
            Optional.ofNullable(cpuOrderString).orElseThrow(() -> new RuntimeException("pod " + lastPodName + " 缓存的 cpu 编号为空。"));
            final String read = FileUtil.read(absolutePath);
            final String replace = read.replace("cpu-order", "\"" + cpuOrderString + "\"");
            FileUtil.writeStringToLocalFile(replace, absolutePath);
            // 创建 pod
            createPod(projectId, yamlPathCacheKey, cpuOrderString);
            log.info("创建项目 " + projectId + " 在节点 " + nodeName + " 的下一个 pod，使用 cpu 编号为 " + cpuOrderString);
        }
    }

    /**
     * @param redisKey yaml 地址的缓存 key
     */
    public void createPodBegin(String projectId, String redisKey) {
        final String podYamlPath = stringRedisTemplate.opsForValue().get(redisKey);
        if (podYamlPath == null) {
            throw new RuntimeException("根据缓存 key 获取 yaml 地址为 null：" + redisKey);
        }
        stringRedisTemplate.delete(redisKey);
        String nodeName = new File(podYamlPath).getName().split("#")[0];
        String podName = podYamlPath.split("#")[1].split("\\.")[0];
        stringRedisTemplate.opsForValue().set("project:" + projectId + ":pod:" + podName + ":node", nodeName);    // 将 pod 运行在哪个 node 上记录到 redis
        new Thread(() -> KubernetesUtil.applyYaml(podYamlPath), "apply-" + podName).start();
    }


    /**
     * @param redisKey yaml 地址的缓存 key
     */
    public void createPod(String projectId, String redisKey, String cpuOrderString) {
        final String podYamlPath = stringRedisTemplate.opsForValue().get(redisKey);
        if (podYamlPath == null) {
            throw new RuntimeException("根据缓存 key 获取 yaml 地址为 null：" + redisKey);
        }
        stringRedisTemplate.delete(redisKey);
        String nodeName = new File(podYamlPath).getName().split("#")[0];
        String podName = podYamlPath.split("#")[1].split("\\.")[0];
        stringRedisTemplate.opsForValue().set("project:" + projectId + ":pod:" + podName + ":node", nodeName);    // 将 pod 运行在哪个 node 上记录到 redis
        stringRedisTemplate.opsForValue().set("project:" + projectId + ":pod:" + podName + ":cpu", cpuOrderString);    // 将 pod 运行在哪个 node 上记录到 redis
        new Thread(() -> KubernetesUtil.applyYaml(podYamlPath), "create-" + podName).start();
    }


    public ProjectEntity getProjectByProjectId(String projectId) {
        ProjectEntity manualProjectEntity = manualProjectMapper.selectById(projectId);
        ProjectEntity autoSubProjectEntity = autoSubProjectMapper.selectById(projectId);
        if (manualProjectEntity != null) {
            return manualProjectEntity;
        } else if (autoSubProjectEntity != null) {
            return autoSubProjectEntity;
        }
        throw new RuntimeException("不存在项目：" + projectId);
    }


//    /**
//     * 获取正在运行的项目的并行度总和
//     *
//     * @param clusterRunningPrefix 集群 key 前缀
//     * @return 正在运行的项目的并行度总和
//     */
//    @SneakyThrows
//    public int getCurrentParallelismSum(String clusterRunningPrefix) {
//        int result = 0;
//        Set<String> clusterRunningKeySet = stringRedisTemplate.keys(clusterRunningPrefix + "*");
//        List<String> runningProjectSet; // 运行中的 projectId 列表
//        if (CollectionUtil.isEmpty(clusterRunningKeySet)) {
//            return 0;
//        }
//        runningProjectSet = getRunningProjectList(clusterRunningKeySet);
//        if (CollectionUtil.isEmpty(runningProjectSet)) {
//            return 0;
//        }
//        for (String projectKey : runningProjectSet) {
//            String projectJsonTemp = stringRedisTemplate.opsForValue().get(projectKey);
//            ProjectMessageDTO projectMessageTemp = JsonUtil.jsonToBean(projectJsonTemp, ProjectMessageDTO.class);
//            result += projectMessageTemp.getCurrentParallelism();   // 获取当前正在使用的并行度
//        }
//        return result;
//    }


    /**
     * 节点剩余可用并行度列表
     *
     * @return 节点映射（节点名，并行度）
     */
    public Map<String, Integer> getNodeMap(String isChoiceGpu) {
        Map<String, Integer> resultNodeMap = new HashMap<>();    // 用于执行的节点映射（节点名，并行度）
        if (DictConstants.USE_GPU.equals(isChoiceGpu)) {
            List<NodeModel> initialNodeList = kubernetesConfiguration.getGpuNodeList();
            log.info("预设并行度的GPU节点列表为：" + initialNodeList);
            for (NodeModel kubernetesNodeSource : initialNodeList) {
                NodeModel kubernetesNodeCopy = kubernetesNodeSource.clone();
                String nodeName = kubernetesNodeCopy.getHostname();
                int maxParallelism = kubernetesNodeCopy.getParallelism();
                String restParallelismKey = "gpu-node:" + nodeName + ":parallelism";
                String restParallelismString = stringRedisTemplate.opsForValue().get(restParallelismKey);
                int restParallelism;
                if (restParallelismString == null) {    // 如果剩余可用并行度没有值，说明是第一次查询，则重置成最大并行度的预设值
                    restParallelism = maxParallelism;
                    stringRedisTemplate.opsForValue().set(restParallelismKey, String.valueOf(restParallelism));
                } else {
                    restParallelism = Integer.parseInt(restParallelismString);
                }
                resultNodeMap.put(nodeName, restParallelism);
            }
            log.info("剩余并行度的GPU节点列表为：" + resultNodeMap);
        } else if (DictConstants.USE_CPU.equals(isChoiceGpu)) {
            List<NodeModel> initialNodeList = kubernetesConfiguration.getCpuNodeList();
            log.info("预设并行度的CPU节点列表为：" + initialNodeList);
            for (NodeModel kubernetesNodeSource : initialNodeList) {
                NodeModel kubernetesNodeCopy = kubernetesNodeSource.clone();
                String nodeName = kubernetesNodeCopy.getHostname();
                int maxParallelism = kubernetesNodeCopy.getParallelism();
                String restParallelismKey = "cpu-node:" + nodeName + ":parallelism";
                String restParallelismString = stringRedisTemplate.opsForValue().get(restParallelismKey);
                int restParallelism;
                if (restParallelismString == null) {    // 如果剩余可用并行度没有值，说明是第一次查询，则重置成最大并行度的预设值
                    restParallelism = maxParallelism;
                    stringRedisTemplate.opsForValue().set(restParallelismKey, String.valueOf(restParallelism));
                } else {
                    restParallelism = Integer.parseInt(restParallelismString);
                }
                resultNodeMap.put(nodeName, restParallelism);
            }
            log.info("剩余并行度的CPU节点列表为：" + resultNodeMap);
        }
        return resultNodeMap;
    }


    /**
     * 根据并行度获取用于执行的节点列表
     * 根据剩余可用并行度降序排序
     *
     * @return 节点映射（节点名，并行度）
     */
    public Map<String, Integer> getNodeMapToUse(String isChoiceGpu, int parallelism) {
        List<NodeModel> initialNodeList; // 预设并行度的节点列表
        if (DictConstants.USE_GPU.equals(isChoiceGpu)) {
            initialNodeList = kubernetesConfiguration.getGpuNodeList();
            log.info("预设并行度的节点列表为：" + initialNodeList);
            // 遍历所有节点，获取还有剩余并行度的节点
            List<NodeModel> restNodeList = new ArrayList<>();    // 剩余并行度的节点列表
            for (NodeModel kubernetesNodeSource : initialNodeList) {
                NodeModel kubernetesNodeCopy = kubernetesNodeSource.clone();
                String nodeName = kubernetesNodeCopy.getHostname();   // 节点名称
                int maxParallelism = kubernetesNodeCopy.getParallelism();
                String restParallelismString = stringRedisTemplate.opsForValue().get("gpu-node:" + nodeName + ":parallelism");// 获取节点剩余并行度的 key
                // -------------------------------- Comment --------------------------------
                int restParallelism;
                if (restParallelismString == null || Integer.parseInt(restParallelismString) > maxParallelism) {    // 如果剩余可用并行度没有值，说明是第一次查询，则重置成最大并行度的预设值
                    restParallelism = maxParallelism;
                    stringRedisTemplate.opsForValue().set("gpu-node:" + nodeName + ":parallelism", String.valueOf(restParallelism));
                } else {
                    restParallelism = Integer.parseInt(restParallelismString);
                    kubernetesNodeCopy.setParallelism(restParallelism);
                }
                if (restParallelism > 0) {
                    restNodeList.add(kubernetesNodeCopy);
                }
            }
            log.info("剩余并行度的节点列表为：" + restNodeList);
            Map<String, Integer> resultNodeMap = new HashMap<>();    // 用于执行的节点映射（节点名，并行度）
            if (!CollectionUtil.isEmpty(restNodeList)) {
                if (restNodeList.size() == 1) {
                    NodeModel tempNode = restNodeList.get(0);
                    String tempNodeName = tempNode.getHostname();
                    int tempParallelism = tempNode.getParallelism();
                    resultNodeMap.put(tempNodeName, Math.min(tempParallelism, parallelism));
                }
                if (restNodeList.size() > 1) {
                    for (int i = 0; i < parallelism; i++) {
                        // 每次降序排序都取剩余并行度最大的一个。
                        restNodeList.sort((o1, o2) -> o2.getParallelism() - o1.getParallelism());
                        NodeModel tempNode = restNodeList.get(0);
                        String tempNodeName = tempNode.getHostname();
                        int tempParallelism = tempNode.getParallelism();
                        if (tempParallelism > 0) {
                            tempNode.setParallelism(tempParallelism - 1);
                            CollectionUtil.addValueToMap(resultNodeMap, 1, tempNodeName);
                        }
                    }
                }
            }
            log.info("即将使用节点的并行度为：" + resultNodeMap);
            return resultNodeMap;
        } else if (DictConstants.USE_CPU.equals(isChoiceGpu)) {
            initialNodeList = kubernetesConfiguration.getCpuNodeList();
            log.info("预设并行度的节点列表为：" + initialNodeList);
            // 遍历所有节点，获取还有剩余并行度的节点
            List<NodeModel> restNodeList = new ArrayList<>();    // 剩余并行度的节点列表
            for (NodeModel kubernetesNodeSource : initialNodeList) {
                NodeModel kubernetesNodeCopy = kubernetesNodeSource.clone();
                String nodeName = kubernetesNodeCopy.getHostname();   // 节点名称
                int maxParallelism = kubernetesNodeCopy.getParallelism();
                String restParallelismString = stringRedisTemplate.opsForValue().get("cpu-node:" + nodeName + ":parallelism");// 获取节点剩余并行度的 key
                // -------------------------------- Comment --------------------------------
                int restParallelism;
                if (restParallelismString == null || Integer.parseInt(restParallelismString) > maxParallelism) {    // 如果剩余可用并行度没有值，说明是第一次查询，则重置成最大并行度的预设值
                    restParallelism = maxParallelism;
                    stringRedisTemplate.opsForValue().set("cpu-node:" + nodeName + ":parallelism", String.valueOf(restParallelism));
                } else {
                    restParallelism = Integer.parseInt(restParallelismString);
                    kubernetesNodeCopy.setParallelism(restParallelism);
                }
                if (restParallelism > 0) {
                    restNodeList.add(kubernetesNodeCopy);
                }
            }
            log.info("剩余并行度的节点列表为：" + restNodeList);
            Map<String, Integer> resultNodeMap = new HashMap<>();    // 用于执行的节点映射（节点名，并行度）
            if (!CollectionUtil.isEmpty(restNodeList)) {
                if (restNodeList.size() == 1) {
                    NodeModel tempNode = restNodeList.get(0);
                    String tempNodeName = tempNode.getHostname();
                    int tempParallelism = tempNode.getParallelism();
                    resultNodeMap.put(tempNodeName, Math.min(tempParallelism, parallelism));
                }
                if (restNodeList.size() > 1) {
                    for (int i = 0; i < parallelism; i++) {
                        // 每次降序排序都取剩余并行度最大的一个。
                        restNodeList.sort((o1, o2) -> o2.getParallelism() - o1.getParallelism());
                        NodeModel tempNode = restNodeList.get(0);
                        String tempNodeName = tempNode.getHostname();
                        int tempParallelism = tempNode.getParallelism();
                        if (tempParallelism > 0) {
                            tempNode.setParallelism(tempParallelism - 1);
                            CollectionUtil.addValueToMap(resultNodeMap, 1, tempNodeName);
                        }
                    }
                }
            }
            log.info("即将使用节点的并行度为：" + resultNodeMap);
            return resultNodeMap;
        } else {
            throw new RuntimeException("未知是否使用 GPU：" + isChoiceGpu);
        }

    }

    /**
     * 获取集群剩余并行度
     *
     * @return 集群剩余并行度
     */
    public int getRestParallelism(String isChoiceGpu) {
        List<NodeModel> initialNodeList; // 预设并行度的节点列表
        if (DictConstants.USE_GPU.equals(isChoiceGpu)) {
            initialNodeList = kubernetesConfiguration.getGpuNodeList(); // 预设并行度的节点列表
            // 遍历所有节点，获取还有剩余并行度的节点
            List<NodeModel> restNodeList = new ArrayList<>();    // 剩余并行度的节点列表
            for (NodeModel kubernetesNodeSource : initialNodeList) {
                NodeModel kubernetesNodeCopy = kubernetesNodeSource.clone();
                String nodeName = kubernetesNodeCopy.getHostname();   // 节点名称
                int maxParallelism = kubernetesNodeCopy.getParallelism();
                String restParallelismString = stringRedisTemplate.opsForValue().get("gpu-node:" + nodeName + ":parallelism");// 获取节点剩余并行度的 key
                // -------------------------------- Comment --------------------------------
                int restParallelism;
                if (restParallelismString == null || Integer.parseInt(restParallelismString) > maxParallelism) {    // 如果剩余可用并行度没有值，说明是第一次查询，则重置成最大并行度的预设值
                    restParallelism = maxParallelism;
                    stringRedisTemplate.opsForValue().set("gpu-node:" + nodeName + ":parallelism", String.valueOf(restParallelism));
                } else {
                    restParallelism = Integer.parseInt(restParallelismString);
                    kubernetesNodeCopy.setParallelism(restParallelism);
                }
                if (restParallelism > 0) {
                    restNodeList.add(kubernetesNodeCopy);
                }
            }
            log.info("集群剩余并行度为：" + restNodeList);
            return restNodeList.size() == 0 ? 0 : restNodeList.stream().mapToInt(NodeModel::getParallelism).sum();
        } else if (DictConstants.USE_CPU.equals(isChoiceGpu)) {
            initialNodeList = kubernetesConfiguration.getCpuNodeList(); // 预设并行度的节点列表
            // 遍历所有节点，获取还有剩余并行度的节点
            List<NodeModel> restNodeList = new ArrayList<>();    // 剩余并行度的节点列表
            for (NodeModel kubernetesNodeSource : initialNodeList) {
                NodeModel kubernetesNodeCopy = kubernetesNodeSource.clone();
                String nodeName = kubernetesNodeCopy.getHostname();   // 节点名称
                int maxParallelism = kubernetesNodeCopy.getParallelism();
                String restParallelismString = stringRedisTemplate.opsForValue().get("cpu-node:" + nodeName + ":parallelism");// 获取节点剩余并行度的 key
                // -------------------------------- Comment --------------------------------
                int restParallelism;
                if (restParallelismString == null || Integer.parseInt(restParallelismString) > maxParallelism) {    // 如果剩余可用并行度没有值，说明是第一次查询，则重置成最大并行度的预设值
                    restParallelism = maxParallelism;
                    stringRedisTemplate.opsForValue().set("cpu-node:" + nodeName + ":parallelism", String.valueOf(restParallelism));
                } else {
                    restParallelism = Integer.parseInt(restParallelismString);
                    kubernetesNodeCopy.setParallelism(restParallelism);
                }
                if (restParallelism > 0) {
                    restNodeList.add(kubernetesNodeCopy);
                }
            }
            log.info("集群剩余并行度为：" + restNodeList);
            return restNodeList.size() == 0 ? 0 : restNodeList.stream().mapToInt(NodeModel::getParallelism).sum();
        } else {
            throw new RuntimeException("未知是否使用 GPU：" + isChoiceGpu);
        }

    }


    public PrefixEntity getRedisPrefixByUserIdAndProjectIdAndTaskId(String userId, String projectId, String taskId) {
        //3 获取用户类型（管理员账户、管理员子账户、普通账户、普通子账户）（独占、共享）
        UserEntity userEntity = userMapper.selectById(userId);
        String roleCode = userEntity.getRoleCode();
        String useType = userEntity.getUseType();
        String clusterId;
        if (DictConstants.ROLE_CODE_SYSADMIN.equals(roleCode) || DictConstants.ROLE_CODE_ADMIN.equals(roleCode)) {  //3-1 管理员账户和管理员子账户直接执行
            clusterId = DictConstants.SYSTEM_CLUSTER_ID;
        } else if (DictConstants.ROLE_CODE_UESR.equals(roleCode)) { //3-2 普通账户，不管是独占还是共享，都在自己的集群里排队，根据自己的独占节点排队
            clusterId = clusterMapper.selectByUserId(userId).getId();
        } else if (DictConstants.ROLE_CODE_SUBUESR.equals(roleCode)) {
            if (DictConstants.USER_TYPE_EXCLUSIVE.equals(useType)) {   //3-3 普通子账户，根据自己的独占节点排队
                clusterId = clusterMapper.selectByUserId(userId).getId();
            } else {    //3-4 共享子账户，根据父账户的共享节点排队
                String parentUserId = userEntity.getCreateUserId();
                clusterId = clusterMapper.selectByUserId(parentUserId).getId();
            }
        } else {
            throw new RuntimeException("未知账户类型，无法获取集群信息。");
        }
        String clusterPrefix = "cluster:" + clusterId;
        String clusterRunningPrefix = clusterPrefix + ":running";
        String projectRunningKey = clusterRunningPrefix + ":" + projectId;
        String taskTickKey = projectRunningKey + ":task:" + taskId + ":tick";
        String taskPodKey = projectRunningKey + ":task:" + taskId + ":pod";
        String taskRetryKey = projectRunningKey + ":task:" + taskId + ":retry";
        String taskMessageKey = projectRunningKey + ":task:" + taskId + ":message";

        return PrefixEntity.builder().clusterPrefix(clusterPrefix).clusterRunningPrefix(clusterRunningPrefix).projectRunningKey(projectRunningKey).taskTickKey(taskTickKey).taskPodKey(taskPodKey).taskRetryKey(taskRetryKey).taskMessageKey(taskMessageKey).build();

    }


    public PrefixEntity getRedisPrefixByClusterIdAndProjectId(String clusterId, String projectId) {
        String clusterPrefix = "cluster:" + clusterId;
        String clusterRunningPrefix = clusterPrefix + ":running";
        String clusterWaitingPrefix = clusterPrefix + ":waiting";
        String projectRunningKey = clusterRunningPrefix + ":" + projectId;
        String projectWaitingKey = clusterWaitingPrefix + ":" + projectId;
        String projectCheckKey = clusterWaitingPrefix + ":" + projectId + ":check";

        return PrefixEntity.builder().clusterPrefix(clusterPrefix).clusterRunningPrefix(clusterRunningPrefix).clusterWaitingPrefix(clusterWaitingPrefix).projectRunningKey(projectRunningKey).projectWaitingKey(projectWaitingKey).projectCheckKey(projectCheckKey).build();

    }


//    /**
//     * 获取 projectId 列表
//     *
//     * @param clusterRunningKeySet 集群下的所有键值对（包括运行中的项目和等待中的项目）
//     * @return projectId 列表
//     */
//    public List<String> getRunningProjectList(Set<String> clusterRunningKeySet) {
//        return clusterRunningKeySet.stream().filter(key -> StringUtil.countSubString(key, ":") == 3).collect(Collectors.toList());
//    }


    public PrefixEntity getRedisPrefixByProjectIdAndProjectType(String projectId, String projectType) {
        String userId;
        if (DictConstants.PROJECT_TYPE_MANUAL.equals(projectType)) {
            userId = manualProjectMapper.selectCreateUserById(projectId);
        } else if (DictConstants.PROJECT_TYPE_AUTO_SUB.equals(projectType)) {
            userId = autoSubProjectMapper.selectCreateUserById(projectId);
        } else {
            throw new RuntimeException("未知的项目类型。");
        }

        //3 获取用户类型（管理员账户、管理员子账户、普通账户、普通子账户）（独占、共享）
        UserEntity userEntity = userMapper.selectById(userId);
        String roleCode = userEntity.getRoleCode();
        String useType = userEntity.getUseType();
        String clusterId;
        if (DictConstants.ROLE_CODE_SYSADMIN.equals(roleCode) || DictConstants.ROLE_CODE_ADMIN.equals(roleCode)) {  //3-1 管理员账户和管理员子账户直接执行
            clusterId = DictConstants.SYSTEM_CLUSTER_ID;
        } else if (DictConstants.ROLE_CODE_UESR.equals(roleCode)) { //3-2 普通账户，不管是独占还是共享，都在自己的集群里排队，根据自己的独占节点排队
            clusterId = clusterMapper.selectByUserId(userId).getId();
        } else if (DictConstants.ROLE_CODE_SUBUESR.equals(roleCode)) {
            if (DictConstants.USER_TYPE_EXCLUSIVE.equals(useType)) {   //3-3 普通子账户，根据自己的独占节点排队
                clusterId = clusterMapper.selectByUserId(userId).getId();
            } else {    //3-4 共享子账户，根据父账户的共享节点排队
                String parentUserId = userEntity.getCreateUserId();
                clusterId = clusterMapper.selectByUserId(parentUserId).getId();
            }
        } else {
            throw new RuntimeException("未知账户类型，无法获取集群信息。");
        }
        String clusterPrefix = "cluster:" + clusterId;
        String clusterRunningPrefix = clusterPrefix + ":running";
        String clusterWaitingPrefix = clusterPrefix + ":waiting";
        String projectRunningKey = clusterRunningPrefix + ":" + projectId;
        String projectWaitingKey = clusterWaitingPrefix + ":" + projectId;
        String projectCheckKey = projectRunningKey + ":check";

        return PrefixEntity.builder().clusterPrefix(clusterPrefix).clusterRunningPrefix(clusterRunningPrefix).clusterWaitingPrefix(clusterWaitingPrefix).projectRunningKey(projectRunningKey).projectWaitingKey(projectWaitingKey).projectCheckKey(projectCheckKey).build();
    }


    public void incrementOneParallelism(String isChoiceGpu, String nodeName) {
        incrementParallelism(isChoiceGpu, nodeName, 1L);
    }

    public void incrementParallelism(String isChoiceGpu, String nodeName, long number) {
        if (DictConstants.USE_GPU.equals(isChoiceGpu)) {
            //1 先检查缓存中的并行度是否超过，超过了就不加缓存的并行度了，常用于测试
            String key = "gpu-node:" + nodeName + ":parallelism";
            final int currentRestParallelism = Integer.parseInt(customRedisClient.get(key));
            final List<NodeModel> nodeList = kubernetesConfiguration.getGpuNodeList();
            for (NodeModel node : nodeList) {
                if (nodeName.equals(node.getHostname())) {
                    if (currentRestParallelism + 1 < node.getParallelism()) {
                        customRedisClient.increment(key, number);
                    }
                }
            }
            log.info("归还 GPU 节点 {} 的 {} 个并行度。", nodeName, number);
        } else if (DictConstants.USE_CPU.equals(isChoiceGpu)) {
            //1 先检查缓存中的并行度是否超过，超过了就不加缓存的并行度了，常用于测试
            String key = "cpu-node:" + nodeName + ":parallelism";
            final int currentRestParallelism = Integer.parseInt(customRedisClient.get(key));
            final List<NodeModel> nodeList = kubernetesConfiguration.getCpuNodeList();
            for (NodeModel node : nodeList) {
                if (nodeName.equals(node.getHostname())) {
                    if (currentRestParallelism + 1 < node.getParallelism()) {
                        customRedisClient.increment(key, number);
                    }
                }
            }
            log.info("归还 CPU 节点 {} 的 {} 个并行度。", nodeName, number);
        }


    }

    public void decrementParallelism(String isChoiceGpu, String nodeName, long number) {
        if (DictConstants.USE_GPU.equals(isChoiceGpu)) {
            String key = "gpu-node:" + nodeName + ":parallelism";
            customRedisClient.decrement(key, number);
            log.info("获取节点 {} 的 {} 个 GPU 并行度。", nodeName, number);
        } else if (DictConstants.USE_CPU.equals(isChoiceGpu)) {
            String key = "cpu-node:" + nodeName + ":parallelism";
            customRedisClient.decrement(key, number);
            log.info("获取节点 {} 的 {} 个 CPU 并行度。", nodeName, number);
        }

    }


    public List<String> getWaitingProjectMessageKeys() {
        final Set<String> keys = stringRedisTemplate.keys("*");
        if (CollectionUtil.isEmpty(keys)) {
            return new ArrayList<>();
        } else {
            return keys.stream().filter(key -> key.contains("cluster") && key.contains("waiting")).collect(Collectors.toList());
        }
    }

    public void resetNodeParallelism() {
        //* -------------------------------- GPU节点 --------------------------------
        final List<NodeModel> gpuNodeList = kubernetesConfiguration.getGpuNodeList();
        log.info("GPU 节点：" + gpuNodeList);
        gpuNodeList.forEach((node) -> customRedisClient.set("gpu-node:" + node.getHostname() + ":parallelism", String.valueOf(node.getParallelism())));
        //* -------------------------------- CPU节点 --------------------------------
        final List<NodeModel> cpuNodeList = kubernetesConfiguration.getCpuNodeList();
        log.info("CPU 节点：" + cpuNodeList);
        cpuNodeList.forEach((node) -> customRedisClient.set("cpu-node:" + node.getHostname() + ":parallelism", String.valueOf(node.getParallelism())));
    }

    /**
     * 判断项目完成度
     */
    @Synchronized
    public boolean complete(PrefixEntity redisPrefix, String projectId) {
        boolean result = false;
        ProjectMessageDTO projectMessageDTO = JsonUtil.jsonToBean(customRedisClient.get(redisPrefix.getProjectRunningKey()), ProjectMessageDTO.class);
        int taskTotal = projectMessageDTO.getTaskTotal();
        int taskCompleted = projectMessageDTO.getTaskCompleted();
        log.info("项目 " + projectId + " 完成进度为：" + (taskCompleted + 1) + "/" + taskTotal);
        if (taskCompleted + 1 == taskTotal) {
            result = true;
        } else {    // 项目没有完成
            projectMessageDTO.setTaskCompleted(taskCompleted + 1);  // 增加已完成任务数
            stringRedisTemplate.opsForValue().set(redisPrefix.getProjectRunningKey(), JsonUtil.beanToJson(projectMessageDTO));
        }
        return result;
    }


    public Integer getUsingLicenseNumber(String userId, String modelType) {
        String key;
        if (DictConstants.MODEL_TYPE_VTD.equals(modelType)) {
            key = "user:" + userId + ":using-license:vtd";
        } else if (DictConstants.MODEL_TYPE_CARSIM.equals(modelType)) {
            key = "user:" + userId + ":using-license:carsim";
        } else {
            throw new RuntimeException("未知模型类型：" + modelType);
        }
        final String usingLicense = customRedisClient.get(key);
        if (StringUtil.isEmpty(usingLicense)) {
            customRedisClient.set(key, "0");
            return 0;
        } else {
            return Integer.parseInt(usingLicense);
        }
    }

    public void useLicense(String userId, String modelType, int parallelism) {
        String key;
        if (DictConstants.MODEL_TYPE_VTD.equals(modelType)) {
            key = "user:" + userId + ":using-license:vtd";
        } else if (DictConstants.MODEL_TYPE_CARSIM.equals(modelType)) {
            key = "user:" + userId + ":using-license:carsim";
        } else {
            throw new RuntimeException("未知模型类型：" + modelType);
        }
        final String usingLicense = customRedisClient.get(key);
        if (StringUtil.isEmpty(usingLicense)) {
            customRedisClient.set(key, String.valueOf(parallelism));
        } else {
            customRedisClient.increment(key, parallelism);
        }
    }


    public void releaseLicense(String projectUserId, String modelType, int parallelism) {
        final String clusterUserId = getClusterUserIdByProjectUserId(projectUserId);
        if (DictConstants.MODEL_TYPE_VTD.equals(modelType)) {
            String key = "user:" + clusterUserId + ":using-license:vtd";
            final String usingLicense = customRedisClient.get(key);
            if (StringUtil.isEmpty(usingLicense)) {
                customRedisClient.set(key, "0");
            } else {
                customRedisClient.decrement(key, parallelism);
            }
        } else if (DictConstants.MODEL_TYPE_CARSIM.equals(modelType)) {
            String key1 = "user:" + clusterUserId + ":using-license:vtd";
            final String usingLicense1 = customRedisClient.get(key1);
            if (StringUtil.isEmpty(usingLicense1)) {
                customRedisClient.set(key1, "0");
            } else {
                customRedisClient.decrement(key1, parallelism);
            }
            String key2 = "user:" + clusterUserId + ":using-license:carsim";
            final String usingLicense2 = customRedisClient.get(key2);
            if (StringUtil.isEmpty(usingLicense2)) {
                customRedisClient.set(key2, "0");
            } else {
                customRedisClient.decrement(key2, parallelism);
            }
        } else {
            throw new RuntimeException("未知模型类型：" + modelType);
        }
    }

    public String getClusterUserIdByProjectUserId(String projectUserId) {
        String clusterUserId;
        final UserEntity userEntity = userMapper.selectById(projectUserId);
        final String roleCode = userEntity.getRoleCode();
        final String useType = userEntity.getUseType();
        if (DictConstants.ROLE_CODE_SYSADMIN.equals(roleCode) || DictConstants.ROLE_CODE_ADMIN.equals(roleCode)) {  //3-1 管理员账户和管理员子账户直接执行
            clusterUserId = projectUserId;
        } else if (DictConstants.ROLE_CODE_UESR.equals(roleCode)) { //3-2 普通账户，不管是独占还是共享，都在自己的集群里排队，根据自己的独占节点排队
            clusterUserId = projectUserId;
        } else if (DictConstants.ROLE_CODE_SUBUESR.equals(roleCode)) {
            if (DictConstants.USER_TYPE_EXCLUSIVE.equals(useType)) {   //3-3 普通子账户，根据自己的独占节点排队
                clusterUserId = projectUserId;
            } else {    //3-4 共享子账户，根据父账户的共享节点排队
                clusterUserId = userEntity.getCreateUserId();
            }
        } else {
            throw new RuntimeException("未知角色类型：" + roleCode);
        }
        return clusterUserId;
    }


    public String getModelTypeByProjectIdAndProjectType(String projectId, String projectType) {
        if (DictConstants.PROJECT_TYPE_MANUAL.equals(projectType)) {
            return manualProjectMapper.selectModelTypeByProjectId(projectId);
        } else if (DictConstants.PROJECT_TYPE_AUTO_SUB.equals(projectType)) {
            return autoSubProjectMapper.selectModelTypeByProjectId(projectId);
        } else {
            throw new RuntimeException("未知项目类型：" + projectType);
        }
    }

    public String getUserIdByProjectIdAndProjectType(String projectId, String projectType) {
        if (DictConstants.PROJECT_TYPE_MANUAL.equals(projectType)) {
            return manualProjectMapper.selectCreateUserById(projectId);
        } else if (DictConstants.PROJECT_TYPE_AUTO_SUB.equals(projectType)) {
            return autoSubProjectMapper.selectCreateUserById(projectId);
        } else {
            throw new RuntimeException("未知项目类型：" + projectType);
        }
    }


    public UserEntity getUserEntityByProjectIdAndProjectType(String projectId, String projectType) {
        UserEntity userEntity;
        if (DictConstants.PROJECT_TYPE_MANUAL.equals(projectType)) {
            userEntity = userMapper.selectByManualProjectId(projectId);
        } else if (DictConstants.PROJECT_TYPE_AUTO_SUB.equals(projectType)) {
            userEntity = userMapper.selectByAutoSubProjectId(projectId);
        } else {
            throw new RuntimeException("未知项目类型：" + projectType);
        }
        return userEntity;
    }


    @Async("pool1")
    @SneakyThrows
    public void selectProjectDetailsById(String projectType, String projectId) {
        String tokenUrl = customConfiguration.getTokenUri() + "?grant_type=client_credentials" + "&client_id=" + customConfiguration.getClientId() + "&client_secret=" + customConfiguration.getClientSecret();
        log.info("获取仿真云平台 token：" + tokenUrl);
        String response = HttpUtil.get(tokenUrl);
        ObjectMapper objectMapper = new ObjectMapper();
        JsonNode jsonNode = objectMapper.readTree(response);
        String accessToken = jsonNode.path("access_token").asText();
        log.info("仿真云平台 token 为：" + accessToken);
        Map<String, String> headers = new HashMap<>();
        headers.put("Authorization", "Bearer " + accessToken);
        Map<String, String> params = new HashMap<>();
        params.put("projectType", projectType);
        params.put("id", projectId);
        String result = HttpUtil.post(closeableHttpClient, requestConfig, customConfiguration.getProjectDetailsUri(), headers, params);
        log.info("访问仿真云平台项目详情接口：" + customConfiguration.getProjectDetailsUri() + "，请求头为：" + headers + "，请求体为：" + params + "结果为：" + result);
        ResponseBodyVO responseBodyVO = JsonUtil.jsonToBean(result, ResponseBodyVO.class);
        String projectDetailsVOJson = JsonUtil.beanToJson(responseBodyVO.getInfo());
        TimeUnit.SECONDS.sleep(5);
        if (DictConstants.PROJECT_TYPE_MANUAL.equals(projectType)) {
            manualProjectMapper.updateProjectDetails(projectId, projectDetailsVOJson);
        } else if (DictConstants.PROJECT_TYPE_AUTO_SUB.equals(projectType)) {
            autoSubProjectMapper.updateProjectDetails(projectId, projectDetailsVOJson);
        } else {
            throw new RuntimeException("未知项目类型：" + projectType);
        }
    }

    @Async("pool1")
    @SneakyThrows
    public void selectProjectReportById(String projectType, String projectId) {
        String tokenUrl = customConfiguration.getTokenUri() + "?grant_type=client_credentials" + "&client_id=" + customConfiguration.getClientId() + "&client_secret=" + customConfiguration.getClientSecret();
        log.info("获取仿真云平台 token：" + tokenUrl);
        String response = HttpUtil.get(tokenUrl);
        ObjectMapper objectMapper = new ObjectMapper();
        JsonNode jsonNode = objectMapper.readTree(response);
        String accessToken = jsonNode.path("access_token").asText();
        log.info("仿真云平台 token 为：" + accessToken);
        Map<String, String> headers = new HashMap<>();
        headers.put("Authorization", "Bearer " + accessToken);
        Map<String, String> params = new HashMap<>();
        params.put("projectType", projectType);
        params.put("id", projectId);
        String result = HttpUtil.post(closeableHttpClient, requestConfig, customConfiguration.getProjectReportUri(), headers, params);
        log.info("访问仿真云平台项目报告接口：" + customConfiguration.getProjectReportUri() + "，请求头为：" + headers + "，请求体为：" + params + "结果为：" + result);
        ResponseBodyVO responseBodyVO = JsonUtil.jsonToBean(result, ResponseBodyVO.class);
        String projectReportVOJson = JsonUtil.beanToJson(responseBodyVO.getInfo());
        TimeUnit.SECONDS.sleep(5);
        if (DictConstants.PROJECT_TYPE_MANUAL.equals(projectType)) {
            manualProjectMapper.updateProjectReport(projectId, projectReportVOJson);
        } else if (DictConstants.PROJECT_TYPE_AUTO_SUB.equals(projectType)) {
            autoSubProjectMapper.updateProjectReport(projectId, projectReportVOJson);
        } else {
            throw new RuntimeException("未知项目类型：" + projectType);
        }
    }
}
